# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, DoubleType, FloatType

from kamae.spark.params import (
    ListwiseStatisticsParams,
    MultiInputSingleOutputParams,
    NanFillValueParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.utils import check_listwise_columns, get_listwise_condition_and_window
from kamae.tensorflow.layers import ListMedianLayer

from .base import BaseTransformer


class ListMedianTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    ListwiseStatisticsParams,
    NanFillValueParams,
):
    """
    Calculate the listwise median across the query id column.
    - If inputCol is set, the transformer calculates the median of the input column
    based on all the items in the same query id column.
    - If inputCols is set, the transformer calculates the median of the first column
    based on second column's topN items in the same query id column.

    By using the topN items to calculate the statistics, we can better approximate
    the real statistics in production. It should be used a large enough topN to get a
    good approximation of the statistics, and an important feature to sort on, such as
    item's production.

    Example: calculate the median price in the same query, based on the top N
    items sorted by descending production.

    :param inputCol: Value column, on which to calculate the median.
    :param inputCols: Input column name.
    - The first is the value column, on which to calculate the median.
    - The second is the sort column, based on which to sort the items.
    :param outputCol: Name of output col.
    :param inputDtype: Data Type of input.
    :param outputDtype: Data Type of output.
    :param layerName: The name of the transformer, which typically
    should be the name of the produced feature.
    :param queryIdCol: Name of column to aggregate upon. It is required.
    :param topN: Filter for limiting the items to calculate the statistics.
    :param sortOrder: Option of 'asc' or 'desc' which defines order
    for listwise operation. Default is 'asc'.
    :param minFilterValue: Minimum value to remove padded values
    defaults to >= 0.
    :nanFillValue: Value to fill NaNs results with. Defaults to 0.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        queryIdCol: Optional[str] = None,
        topN: Optional[int] = None,
        sortOrder: str = "asc",
        minFilterValue: Optional[float] = None,
        nanFillValue: float = 0.0,
    ) -> None:
        super().__init__()
        self._setDefault(
            topN=None,
            sortOrder="asc",
            minFilterValue=None,
            nanFillValue=0,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Calculate the listwise median, optionally sorting and
        filtering based on the second input column.
        :param dataset: The dataframe with signals and features.
        :returns: The dataframe dataset with the new feature.
        """
        if not self.isDefined("queryIdCol"):
            raise ValueError("queryIdCol must be set on listwise transformers.")

        # Define the columns to use for the calculation
        with_sort = self.isDefined("inputCols")
        if with_sort:
            val_col_name = self.getInputCols()[0]
            sort_col_name = self.getInputCols()[1]
        else:
            val_col_name = self.getInputCol()
            sort_col_name = None

        check_listwise_columns(
            dataset=dataset,
            query_col_name=self.getQueryIdCol(),
            value_col_name=val_col_name,
            sort_col_name=sort_col_name,
        )

        cond_col, window_spec = get_listwise_condition_and_window(
            query_col=F.col(self.getQueryIdCol()),
            value_col=F.col(val_col_name),
            sort_col=F.col(sort_col_name) if with_sort else None,
            sort_order=self.getSortOrder(),
            sort_top_n=self.getTopN(),
            min_filter_value=self.getMinFilterValue(),
        )

        # Calculate the statistics under the conditions
        dataset = dataset.withColumn(
            "sorted_values",
            F.sort_array(
                F.collect_list(F.when(cond_col, F.col(val_col_name))).over(window_spec)
            ),
        )

        # Compute median using Spark native functions
        dataset = dataset.withColumn("array_size", F.size(F.col("sorted_values")))
        mid_index_1 = (F.col("array_size") / 2).cast("int")
        mid_index_2 = (F.col("array_size") / 2 - 1).cast("int")
        dataset = dataset.withColumn(
            self.getOutputCol(),
            F.when(
                F.col("array_size") % 2 == 1,
                F.element_at(F.col("sorted_values"), mid_index_1 + 1),
            ).otherwise(
                (
                    F.element_at(F.col("sorted_values"), mid_index_1 + 1)
                    + F.element_at(F.col("sorted_values"), mid_index_2 + 1)
                )
                / 2.0
            ),
        )

        # Replace Nulls/Nans
        dataset = dataset.fillna({self.getOutputCol(): self.getNanFillValue()})

        return dataset

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the listwise-median transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a median operation.
        """
        return ListMedianLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            top_n=self.getTopN(),
            sort_order=self.getSortOrder(),
            min_filter_value=self.getMinFilterValue(),
            nan_fill_value=self.getNanFillValue(),
            axis=1,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple


class InputOutputExtractor:
    """
    Mixin class containing methods for extracting input and output column names.
    """

    def get_layer_inputs_outputs(self) -> Tuple[List[str], List[str]]:
        """
        Gets the input & output information of the layer. Returns a tuple of lists,
        the first containing the input column names and the second containing the
        output column names.

        :returns: Tuple of lists containing the input and output column names.
        """

        if hasattr(self, "input_cols") and getattr(self, "input_cols") is not None:
            inputs = self.input_cols
        elif hasattr(self, "input_col") and getattr(self, "input_col") is not None:
            inputs = [self.input_col]
        else:
            inputs = []

        if hasattr(self, "output_cols") and getattr(self, "output_cols") is not None:
            outputs = self.output_cols
        elif hasattr(self, "output_col") and getattr(self, "output_col") is not None:
            outputs = [self.output_col]
        else:
            outputs = []

        return inputs, outputs



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import Hashing

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class HashIndexLayer(BaseLayer):
    """
    Wrapper around the Keras Hashing layer which hashes and bins categorical features.

    This layer transforms categorical inputs to hashed output. It element-wise
    converts ints or strings to ints in a fixed range. The stable hash
    function uses `tensorflow::ops::Fingerprint` to produce the same output
    consistently across all platforms.

    This layer uses [FarmHash64](https://github.com/google/farmhash),
    which provides a consistent hashed output across different platforms and is
    stable across invocations, regardless of device and context, by mixing the
    input bits thoroughly.
    """

    def __init__(
        self,
        num_bins: int,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        mask_value: Union[int, str] = None,
        **kwargs,
    ) -> None:
        """
        Intialise the HashIndexLayer layer.

        :param num_bins: Number of hash bins. Note that this includes the `mask_value`
        bin, so the effective number of bins is `(num_bins - 1)` if `mask_value`
        is set.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param mask_value: A value that represents masked inputs, which are mapped to
        index 0. Defaults to None, meaning no mask term will be added and the
        hashing will start at index 0.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.num_bins = num_bins
        self.mask_value = mask_value
        self.hash_indexer = Hashing(name=name, num_bins=num_bins, mask_value=mask_value)

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the hash indexing on the input tensor by calling the underlying
        Hashing layer.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to be hashed.
        :returns: Hashed and bucketed tensor.
        """
        return self.hash_indexer(inputs)

    def get_config(self) -> Dict[str, Any]:
        """
        Returns the configuration of the HashIndexLayer layer.

        :returns: Configuration of the HashIndexLayer layer.
        """
        config = super().get_config()
        config.update({"num_bins": self.num_bins, "mask_value": self.mask_value})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, IntegerType, LongType, ShortType, StringType

from kamae.spark.params import (
    DropUnseenParams,
    MultiInputMultiOutputParams,
    StringIndexParams,
)
from kamae.spark.transformers import SharedOneHotEncodeTransformer
from kamae.spark.utils import collect_labels_array_from_multiple_columns

from .base import BaseEstimator


class SharedOneHotEncodeEstimator(
    BaseEstimator,
    MultiInputMultiOutputParams,
    DropUnseenParams,
    StringIndexParams,
):
    """
    Shared One-hot encoder Spark Estimator for use in Spark pipelines.
    This estimator is used to collect all the string labels for a given set of input
    columns. When fit is called it returns a SharedOneHotEncodeTransformer which can be
    used to create one-hot arrays from additional feature columns using the
    same string labels.
    """

    @keyword_only
    def __init__(
        self,
        inputCols: Optional[List[str]] = None,
        outputCols: Optional[List[str]] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        stringOrderType: str = "frequencyDesc",
        maskToken: Optional[str] = None,
        numOOVIndices: int = 1,
        dropUnseen: bool = False,
        maxNumLabels: Optional[int] = None,
    ) -> None:
        """
        Initializes the SharedOneHotEncoder estimator.
        Sets all parameters to given inputs.

        :param inputCols: List of input column names.
        :param outputCols: List of output column names.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer returned
        by the fit method.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column(s) to after
        transforming.
        :param: stringOrderType: How to order the string indices.
        Options are 'frequencyAsc', 'frequencyDesc', 'alphabeticalAsc',
        'alphabeticalDesc'. Defaults to 'frequencyDesc'.
        :param maskToken: Token to use for masking.
        If set, the token will be indexed as 0.
        :param numOOVIndices: Number of out of vocabulary indices to use. The
        out of vocabulary indices are used to represent unseen labels and are
        placed at the beginning of the one-hot encoding. Defaults to 1.
        :param dropUnseen: Whether to drop unseen label indices. If set to True,
        the transformer will not add an extra dimension for unseen labels in the
        one-hot encoding. Defaults to False.
        :param maxNumLabels: Optional value to limit the size of the vocabulary.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            stringOrderType="frequencyDesc",
            numOOVIndices=1,
            dropUnseen=False,
            maskToken=None,
            maxNumLabels=None,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [ShortType(), IntegerType(), LongType(), StringType()]

    def _fit(self, dataset: DataFrame) -> "SharedOneHotEncodeTransformer":
        """
        Fits the OneHotEncodeEstimator estimator to the given dataset.
        Returns a OneHotEncodeTransformer which can be used to one-hot columns using
        the collected string labels.

        :param dataset: Pyspark dataframe to fit the estimator to.
        :returns: OneHotEncodeTransformer instance with collected string labels.
        """
        column_datatypes = [
            self.get_column_datatype(dataset=dataset, column_name=i)
            for i in self.getInputCols()
        ]
        labels = collect_labels_array_from_multiple_columns(
            dataset=dataset,
            columns=[F.col(i) for i in self.getInputCols()],
            column_datatypes=column_datatypes,
            string_order_type=self.getStringOrderType(),
            mask_token=self.getMaskToken(),
            max_num_labels=self.getMaxNumLabels(),
        )
        self.setLabelsArray(labels)

        return SharedOneHotEncodeTransformer(
            inputCols=self.getInputCols(),
            outputCols=self.getOutputCols(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            labelsArray=self.getLabelsArray(),
            stringOrderType=self.getStringOrderType(),
            maskToken=self.getMaskToken(),
            numOOVIndices=self.getNumOOVIndices(),
            dropUnseen=self.getDropUnseen(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from pyspark.ml import Estimator
from pyspark.sql import DataFrame

from kamae.spark.common import SparkOperation
from kamae.spark.transformers import BaseTransformer

if TYPE_CHECKING:
    from pyspark.ml._typing import ParamMap


class BaseEstimator(Estimator, SparkOperation):
    def __init__(self):
        """
        Initializes the estimator.
        """
        super().__init__()

    def fit(
        self,
        dataset: DataFrame,
        params: Optional[Union["ParamMap", List["ParamMap"], Tuple["ParamMap"]]] = None,
    ) -> BaseTransformer:
        """
        Overrides the fit method of the parent class to add casting of input columns
        to the preferred data type.

        :param dataset: Input dataset.
        :param params: Optional additional parameters.
        :returns: Fitted transformer companion object.
        """
        try:
            dataset = self._create_casted_input_output_columns(
                dataset=dataset, ingress=True
            )
            self._check_input_dtypes_compatible(
                dataset, self._get_single_or_multi_col(ingress=True)
            )

            # Set estimator input columns to casted columns
            self.set_input_columns_to_from_casted(
                dataset=dataset,
                suffix=self.tmp_column_suffix,
            )

            # Replicate the logic from the existing abstract estimator fit method
            transformer = super().fit(dataset, params)

            # Reset input columns from casted columns
            self.set_input_columns_to_from_casted(
                dataset=dataset,
                suffix=self.tmp_column_suffix,
                reverse=True,
            )

            # Reset input columns for transformer
            transformer.set_input_columns_to_from_casted(
                dataset=dataset,
                suffix=self.tmp_column_suffix,
                reverse=True,
            )

            return transformer

        except Exception as e:
            param_dict = {
                param[0].name: param[1] for param in self.extractParamMap().items()
            }
            raise e.__class__(
                f"Error in estimator: {self.uid} with params: {param_dict}"
            ).with_traceback(e.__traceback__)

    def construct_layer_info(self) -> Dict[str, Any]:
        """
        Constructs the layer info dictionary.
        Contains the layer name, the tensorflow layer, and the inputs and outputs.
        This is used when constructing the pipeline graph.

        layer is set to None because estimators do not have a defined tf computation,
        we will not use this information in the pipeline graph (for estimators).

        :returns: Dictionary containing layer information such as
        name, tensorflow layer, inputs, and outputs.
        """
        inputs, outputs = self.get_layer_inputs_outputs()
        return {
            "name": self.getOrDefault("layerName"),
            # Estimators do not have a defined tf computation, only transformers do.
            "layer": None,
            "inputs": inputs,
            "outputs": outputs,
        }



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from .name import LayerNameMixin


class SingleInputMixin:
    """
    Mixin class containing set methods for the single input column scenario.
    """

    _input_col: str

    @property
    def input_col(self) -> str:
        """
        Gets the input column name.

        :returns: Input column name.
        """
        return self._input_col

    @input_col.setter
    def input_col(self, value: str) -> None:
        """
        Sets the input column name.

        :param value: String to set the input_col parameter to.
        :returns: None, input_col is set to the given value.
        """
        self._input_col = value


class MultiInputMixin:
    """
    Mixin class containing set methods for the multiple input columns scenario.
    """

    _input_cols: List[str]

    @property
    def input_cols(self) -> List[str]:
        """
        Gets the input column names.

        :returns: List of strings of input column names.
        """
        return self._input_cols

    @input_cols.setter
    def input_cols(self, value: List[str]) -> None:
        """
        Sets the input column names. to the given list of strings.

        :param value: List of strings to set the input_col parameter to.
        :returns: None, input_col is set to the given value.
        """
        self._input_cols = value


class SingleOutputMixin(LayerNameMixin):
    """
    Mixin class containing set methods for the single output column scenario.
    """

    _output_col: str

    @property
    def output_col(self) -> str:
        """
        Gets the output column name.

        :returns: List of strings of output column names.
        """
        return self._output_col

    @output_col.setter
    def output_col(self, value: str) -> None:
        """
        Sets the output column name to the given string value.
        Throws an error if the value is the same as the layer name,
        as this causes issues when constructing the pipeline graph.

        :param value: String to set the output_col parameter to.
        :returns: None, output_col is set to the given value.
        """
        if value is None:
            # Set default output column name
            self._output_col = "output"
        if hasattr(self, "layer_name") and self.layer_name == value:
            raise ValueError(
                f"""Output column name {value} cannot be the same
                as the layer name {self.layer_name}"""
            )
        self._output_col = value

    @LayerNameMixin.layer_name.setter
    def layer_name(self, value: str) -> None:
        """
        Sets the layer name to the given string value.
        Throws an error if the value is the same as the output column name,
        as this causes issues when constructing the pipeline graph.

        :param value: String to set the layer_name parameter to.
        :returns: None, layer_name is set to the given value.
        """
        if hasattr(self, "output_col") and self.output_col == value:
            raise ValueError(
                f"""Layer name {value} cannot be the same
                as the output column name {self.output_col}"""
            )
        self._layer_name = value if value is not None else self.__repr__()


class MultiOutputMixin(LayerNameMixin):
    """
    Mixin class containing set methods for the multiple output columns scenario.
    """

    _output_cols: List[str]

    @property
    def output_cols(self) -> List[str]:
        """
        Gets the output column names.

        :returns: List of strings of output column names.
        """
        return self._output_cols

    @LayerNameMixin.layer_name.setter
    def layer_name(self, value: str) -> None:
        """
        Sets the layer name to the given string value.
        Throws an error if the value is the same as any of the output column names,
        as this causes issues when constructing the pipeline graph.

        :param value: String to set the layer_name parameter to.
        :returns: None, layer_name is set to the given value.
        """
        if hasattr(self, "output_cols") and any(
            [output_col == value for output_col in self.output_cols]
        ):
            raise ValueError(
                f"""Layer name {value} cannot be the same
                as any of the output column names {", ".join(self.output_cols)}"""
            )
        self._layer_name = value if value is not None else self.__repr__()

    @output_cols.setter
    def output_cols(self, value: List[str]) -> None:
        """
        Sets the output column names to the given list of strings.
        Throws an error if any of the values in the list is the same as the layer name,
        as this causes issues when constructing the pipeline graph.

        :param value: List of strings to set the output_cols parameter to.
        :returns: None, output_cols is set to the given value.
        """
        if hasattr(self, "layer_name") and self.layer_name in value:
            raise ValueError(
                f"""Output column names {", ".join(value)} cannot contain
                the layer name {self.layer_name}"""
            )
        self._output_cols = value


class SingleInputSingleOutputMixin(SingleInputMixin, SingleOutputMixin):
    """
    Mixin for a layer that takes a single input and returns a single output
    """


class SingleInputMultiOutputMixin(SingleInputMixin, MultiOutputMixin):
    """
    Mixin for a layer that takes a single input and returns multiple outputs
    """


class MultiInputSingleOutputMixin(MultiInputMixin, SingleOutputMixin):
    """
    Mixin for a layer that takes multiple inputs and returns a single output
    """


class MultiInputMultiOutputMixin(MultiInputMixin, MultiOutputMixin):
    """
    Mixin for a layer that takes multiple inputs and returns multiple outputs
    """



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class CosineSimilarityLayer(BaseLayer):
    """
    Computes the cosine similarity between two input tensors.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        axis: int = -1,
        keepdims: bool = False,
        **kwargs,
    ) -> None:
        """
        Initializes the CosineSimilarityLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: The axis along which to compute the cosine similarity. Defaults to
        `-1`.
        :param keepdims: Whether to keep the shape of the input tensor. Defaults to
        `False`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.axis = axis
        self.keepdims = keepdims

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.complex64,
            tf.complex128,
        ]

    @enforce_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs) -> Tensor:
        """
        Computes the cosine similarity between two input tensors. If `keepdims` is
        `True`, the shape is retained. Otherwise, the shape is reduced along the
        specified axis.

        Decorated with @enforce_multiple_tensor_input to ensure that the input
        is an iterable of tensors. Raises an error if a single tensor is passed.

        After decoration, we check the length of the inputs to ensure we have the right
        number of input tensors.

        :param inputs: List of two tensors to compute the cosine similarity between.
        :returns: The tensor resulting from the cosine similarity.
        """
        if len(inputs) != 2:
            raise ValueError(
                f"Expected 2 inputs, received {len(inputs)} inputs instead."
            )
        x = tf.nn.l2_normalize(inputs[0], axis=self.axis)
        y = tf.nn.l2_normalize(inputs[1], axis=self.axis)

        return tf.reduce_sum(tf.multiply(x, y), axis=self.axis, keepdims=self.keepdims)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the CosineSimilarity layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"axis": self.axis, "keepdims": self.keepdims})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List, Optional, Type

import networkx as nx
from pyspark import keyword_only
from pyspark.ml import Pipeline
from pyspark.ml.param import Params
from pyspark.ml.pipeline import PipelineReader, PipelineSharedReadWrite, PipelineWriter
from pyspark.ml.util import DefaultParamsReader, MLWriter
from pyspark.sql import DataFrame

from kamae.graph import PipelineGraph
from kamae.spark.estimators import BaseEstimator
from kamae.spark.pipeline import KamaeSparkPipelineModel
from kamae.spark.transformers import BaseTransformer

if TYPE_CHECKING:
    from pyspark.ml._typing import ParamMap

    from kamae.spark.pipeline import KamaePipelineStage


class KamaeSparkPipeline(Pipeline):
    """
    KamaeSparkPipeline is a subclass of pyspark.ml.Pipeline that is used to chain
    together BaseTransformers.
    It maintains the same functionality as pyspark.ml.Pipeline e.g. serialisation.
    """

    @keyword_only
    def __init__(self, *, stages: Optional[List["KamaePipelineStage"]] = None) -> None:
        """
        Initialises the KamaeSparkPipeline object.

        :param stages: List of LayerTransformers to chain together.
        :returns: None - class instantiated.
        """
        super().__init__(stages=stages)

    def setStages(self, value: List["KamaePipelineStage"]) -> "KamaeSparkPipeline":
        """
        Sets the stages of the pipeline.

        :param value: List of pipeline stages.
        :returns: KamaeSparkPipeline object with stages set.
        """
        return self._set(stages=value)

    def getStages(self) -> List["KamaePipelineStage"]:
        """
        Gets the stages of the pipeline.

        :returns: List of pipeline stages.
        """
        return self.getOrDefault("stages")

    @keyword_only
    def setParams(
        self, *, stages: Optional["KamaePipelineStage"] = None
    ) -> "KamaeSparkPipeline":
        """
        Sets the keyword arguments of the pipeline.

        :param stages: List of pipeline stages.
        :returns: KamaeSparkPipeline object with stages set.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def expand_pipeline_stages(self) -> List["KamaePipelineStage"]:
        """
        Expands the pipeline stages to include all nested pipeline stages.
        If the pipeline stage is itself a pipeline model, it will be expanded
        recursively.

        :returns: List of all pipeline stages flattened to transformer level.
        """
        expanded_stages = []
        for stage in self.getStages():
            if isinstance(stage, (KamaeSparkPipelineModel, KamaeSparkPipeline)):
                # Recursively expand the pipeline stages.
                expanded_stages.extend(stage.expand_pipeline_stages())
            else:
                expanded_stages.append(stage)
        return expanded_stages

    @staticmethod
    def collect_estimator_parents(
        stages: List["KamaePipelineStage"],
    ) -> List["KamaePipelineStage"]:
        """
        Collects the parent stages of the estimators in the pipeline.

        Used to determine which transformers to execute before the estimators in the
        pipeline.

        :param stages: List of pipeline stages.
        :returns: List of names of the ancestors of the estimators in the pipeline.
        """
        stage_dict = {
            stage.getOrDefault("layerName"): stage.construct_layer_info()
            for stage in stages
        }
        pipeline_graph = PipelineGraph(stage_dict=stage_dict)
        estimator_stages = [
            stage for stage in stages if isinstance(stage, BaseEstimator)
        ]
        estimator_parents = []
        for estimator in estimator_stages:
            layer_name = estimator.getLayerName()
            specific_estimator_parents = nx.ancestors(pipeline_graph.graph, layer_name)
            estimator_parents.extend(specific_estimator_parents)

        distinct_estimator_parents = list(set(estimator_parents))
        estimator_parent_stages = [
            stage
            for stage in stages
            if stage.getLayerName() in distinct_estimator_parents
        ]
        return estimator_parent_stages

    def _fit(self, dataset: DataFrame) -> "KamaeSparkPipelineModel":
        """
        Fits the pipeline to the dataset. Returns a KamaeSparkPipelineModel object.

        Calls the super fit method of the pyspark.ml.Pipeline class and
        then constructs a KamaeSparkPipelineModel uses the stages from the fit pipeline.

        :param dataset: PySpark DataFrame to fit the pipeline to.
        :returns: KamaeSparkPipelineModel object.
        """
        expanded_pipeline_stages = self.expand_pipeline_stages()

        for stage in expanded_pipeline_stages:
            if not (
                isinstance(stage, BaseEstimator) or isinstance(stage, BaseTransformer)
            ):
                raise TypeError(
                    "Cannot recognize a pipeline stage of type %s." % type(stage)
                )

        # Native Spark checks for the last estimator and executes all transformers
        # before it, regardless whether there is a dependency between them. See here:
        # https://github.com/apache/spark/blob/master/python/pyspark/ml/pipeline.py#L120
        # We can be clever, since we have built a proper DAG, by only executing
        # transformers that are required by the estimator.

        # Collect the parents of the estimators in the pipeline
        estimator_parent_stages = self.collect_estimator_parents(
            expanded_pipeline_stages
        )
        # Fit each stage, appending the transformer to the list of transformers
        # If the stage is a parent of an estimator, transform the dataset.
        transformers: List[BaseTransformer] = []
        for stage in expanded_pipeline_stages:
            if isinstance(stage, BaseTransformer):
                transformers.append(stage)
                if stage in estimator_parent_stages:
                    dataset = stage.transform(dataset)
            else:
                model = stage.fit(dataset)
                transformers.append(model)
                if stage in estimator_parent_stages:
                    dataset = model.transform(dataset)
        return KamaeSparkPipelineModel(transformers)

    def copy(self, extra: Optional["ParamMap"] = None) -> "KamaeSparkPipeline":
        """
        Creates a copy of the KamaeSparkPipeline object.

        :param extra: Additional optional params to copy to new pipeline.
        :returns: KamaeSparkPipeline object.
        """
        if extra is None:
            extra = dict()
        that = Params.copy(self, extra)
        stages = [stage.copy(extra) for stage in that.getStages()]
        return that.setStages(stages)

    def write(self) -> MLWriter:
        """
        Uses the KamaeSparkPipelineWriter class to write the pipeline to a
        persistent storage path.

        :returns: KamaeSparkPipelineWriter object.
        """
        return KamaeSparkPipelineWriter(self)

    @classmethod
    def read(cls) -> "KamaeSparkPipelineReader":
        """
        Uses the KamaeSparkPipelineReader class to read a pipeline from a
        persistent storage path.

        :returns: KamaeSparkPipelineReader object.
        """
        return KamaeSparkPipelineReader(cls)


class KamaeSparkPipelineReader(PipelineReader):
    """
    Util class for reading a pipeline from a persistent storage path.
    """

    def __init__(self, cls: Type[KamaeSparkPipeline]):
        super().__init__(cls=cls)

    def load(self, path: str) -> KamaeSparkPipeline:
        """
        Loads a pipeline from a given path.

        :param path: Path to stored pipeline.
        :returns: KamaeSparkPipeline object.
        """
        metadata = DefaultParamsReader.loadMetadata(path, self.sc)
        uid, stages = PipelineSharedReadWrite.load(metadata, self.sc, path)
        return KamaeSparkPipeline(stages=stages)._resetUid(uid)


class KamaeSparkPipelineWriter(PipelineWriter):
    """
    Util class for writing a pipeline to a persistent storage path.
    """

    def __init__(self, instance: KamaeSparkPipeline):
        super().__init__(instance=instance)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

from pyspark.ml.param import Param, Params, TypeConverters


class StringConstantParams(Params):
    """
    Mixin class for a string constant.
    """

    stringConstant = Param(
        Params._dummy(),
        "stringConstant",
        "String constant to use in many string transformers",
        typeConverter=TypeConverters.toString,
    )

    def setStringConstant(self, value: str) -> "StringConstantParams":
        """
        Sets the stringConstant parameter.

        :param value: String constant value to use in different string transformers.
        :returns: Instance of class mixed in.
        """
        return self._set(stringConstant=value)

    def getStringConstant(self) -> str:
        """
        Gets the stringConstant parameter.

        :returns: String constant value to use in different string transformers.
        """
        return self.getOrDefault(self.stringConstant)


class NegationParams(Params):
    """
    Mixin class containing negation parameter needed for transforms that output a
    boolean.
    """

    negation = Param(
        Params._dummy(),
        "negation",
        "Whether to negate the operation.",
        typeConverter=TypeConverters.toBoolean,
    )

    def setNegation(self, value: bool) -> "NegationParams":
        """
        Sets the negation parameter.

        :param value: Bool value of whether to negate the operation.
        :returns: Instance of class mixed in.
        """
        return self._set(negation=value)

    def getNegation(self) -> bool:
        """
        Gets the negation parameter.

        :returns: Bool value of whether to negate the operation.
        """
        return self.getOrDefault(self.negation)


class MathFloatConstantParams(Params):
    """
    Mixin class for a math float constant.
    """

    mathFloatConstant = Param(
        Params._dummy(),
        "mathFloatConstant",
        "Float constant used for math operations",
        typeConverter=TypeConverters.toFloat,
    )

    def __init__(self) -> None:
        super().__init__()
        self._setDefault(layerName=self.uid)

    def getMathFloatConstant(self) -> float:
        """
        Gets the value of the mathFloatConstant parameter.

        :returns: Float constant used for math operations.
        """
        return self.getOrDefault(self.mathFloatConstant)

    def setMathFloatConstant(self, value: float) -> "MathFloatConstantParams":
        """
        Sets the value of the mathFloatConstant parameter.

        :param value: Float constant used for math operations.
        :returns: Class instance.
        """
        return self._set(mathFloatConstant=value)


class StringRegexParams(Params):
    """
    Mixin class for string transformers that use regex.
    """

    regex = Param(
        Params._dummy(),
        "regex",
        "Whether to use regex in the string contains operation.",
        typeConverter=TypeConverters.toBoolean,
    )

    def setRegex(self, value: bool) -> "StringRegexParams":
        """
        Sets the regex parameter.

        :param value: Bool value of whether to use regex in the string contains
        operation.
        :returns: Instance of class mixed in.
        """
        return self._set(regex=value)

    def getRegex(self) -> bool:
        """
        Gets the regex parameter.

        :returns: Bool value of whether to use regex in the string contains
         operation.
        """
        return self.getOrDefault(self.regex)


class ConstantStringArrayParams(Params):
    """
    Mixin class containing separator parameter needed for constant string array
    transforms.
    """

    constantStringArray = Param(
        Params._dummy(),
        "constantStringArray",
        "Value to fill the column with.",
        typeConverter=TypeConverters.toListString,
    )

    def setConstantStringArray(self, value: List[str]) -> "ConstantStringArrayParams":
        """
        Sets the constantStringArray parameter.

        :param value: List of strings to use as a constant string array.
        :returns: Instance of class mixed in.
        """
        return self._set(constantStringArray=value)

    def getConstantStringArray(self) -> List[str]:
        """
        Gets the constantStringArray parameter.

        :returns: List of strings to use as a constant string array.
        """
        return self.getOrDefault(self.constantStringArray)


class LabelsArrayParams(Params):
    labelsArray = Param(
        Params._dummy(),
        "labelsArray",
        "Ordered list of labels to use for the indexer",
        typeConverter=TypeConverters.toListString,
    )

    def setLabelsArray(self, value: List[str]) -> "LabelsArrayParams":
        """
        Sets the labelArray parameter.

        :param value: List of strings to use in indexing transformers.
        :returns: Instance of class mixed in.
        """
        return self._set(labelsArray=value)

    def getLabelsArray(self) -> List[str]:
        """
        Gets the labelArray parameter.

        :returns: List of strings to use in indexing transformers.
        """
        return self.getOrDefault(self.labelsArray)


class DropUnseenParams(Params):
    """
    Mixin class containing parameters needed to drop unseen index in the
    one hot encoder layer.
    """

    dropUnseen = Param(
        Params._dummy(),
        "dropUnseen",
        "Whether to the drop unseen label index in the one hot encoder layer",
        typeConverter=TypeConverters.toBoolean,
    )

    def setDropUnseen(self, value: bool) -> "DropUnseenParams":
        """
        Sets the dropUnseen parameter.

        :param value: Bool value of whether to drop unseen label index
        in the one hot encoder layer.
        :returns: Instance of class mixed in.
        """
        return self._set(dropUnseen=value)

    def getDropUnseen(self) -> bool:
        """
        Gets the dropUnseen parameter.

        :returns: Bool value of whether to drop unseen label index
        in the one hot encoder layer.
        """
        return self.getOrDefault(self.dropUnseen)


class StringIndexParams(LabelsArrayParams):
    """
    Mixin class containing parameters needed for string indexer and one hot encoder
    layers.
    """

    stringOrderType = Param(
        Params._dummy(),
        "stringOrderType",
        """How to order the strings. Options are
        'frequencyAsc', 'frequencyDesc', 'alphabeticalAsc', 'alphabeticalDesc'.""",
        typeConverter=TypeConverters.toString,
    )

    maskToken = Param(
        Params._dummy(),
        "maskToken",
        "Mask token to use for string indexing",
        typeConverter=TypeConverters.toString,
    )

    numOOVIndices = Param(
        Params._dummy(),
        "numOOVIndices",
        "Number of out of vocabulary indices to use",
        typeConverter=TypeConverters.toInt,
    )

    maxNumLabels = Param(
        Params._dummy(),
        "maxNumLabels",
        "Max number of labels to use",
        typeConverter=TypeConverters.toInt,
    )

    def setStringOrderType(self, value: str) -> "StringIndexParams":
        """
        Sets the stringOrderType parameter to the given value.
        Must be one of:
        - 'frequencyAsc'
        - 'frequencyDesc'
        - 'alphabeticalAsc'
        - 'alphabeticalDesc'.

        :param value: String to set the stringOrderType parameter to.
        :returns: Instance of class mixed in.
        """
        possible_order_options = [
            "frequencyAsc",
            "frequencyDesc",
            "alphabeticalAsc",
            "alphabeticalDesc",
        ]
        if value not in possible_order_options:
            raise ValueError(
                f"stringOrderType must be one of {', '.join(possible_order_options)}"
            )
        return self._set(stringOrderType=value)

    def getStringOrderType(self) -> str:
        """
        Gets the stringOrderType parameter.

        :returns: String value of the stringOrderType parameter.
        """
        return self.getOrDefault(self.stringOrderType)

    def setMaskToken(self, value: str) -> "StringIndexParams":
        """
        Sets the maskToken parameter.

        :param value: String value for the mask token to use for string indexing.
        :returns: Instance of class mixed in.
        """
        return self._set(maskToken=value)

    def getMaskToken(self) -> str:
        """
        Gets the maskToken parameter.

        :returns: String value for the mask token to use for string indexing.
        """
        return self.getOrDefault(self.maskToken)

    def setNumOOVIndices(self, value: int) -> "StringIndexParams":
        """
        Sets the numOOVIndices parameter.

        :param value: Int value for the number of OOV indices to use for string
        indexing.
        :returns: Instance of class mixed in.
        """
        if value <= 0:
            raise ValueError("numOOVIndices must be a positive integer")
        return self._set(numOOVIndices=value)

    def getNumOOVIndices(self) -> int:
        """
        Gets the numOOVIndices parameter.

        :returns: Int value for the number of OOV indices to use for string
        indexing.
        """
        return self.getOrDefault(self.numOOVIndices)

    def setMaxNumLabels(self, value: int) -> "StringIndexParams":
        """
        Sets the maxNumLabels parameter.

        :param value: Int value for the max number of labels to use for string
        indexing.
        :returns: Instance of class mixed in.
        """
        if value <= 0:
            raise ValueError("maxNumLabels must be a positive integer")
        return self._set(maxNumLabels=value)

    def getMaxNumLabels(self) -> int:
        """
        Gets the maxNumLabels parameter.

        :returns: Int value for the max number of labels to use for string
        indexing.
        """
        return self.getOrDefault(self.maxNumLabels)


class HashIndexParams(Params):
    """
    Mixin class containing bin parameter needed for hash indexing layers.
    """

    numBins = Param(
        Params._dummy(),
        "numBins",
        "Number of bins to use for hash indexing",
        typeConverter=TypeConverters.toInt,
    )

    maskValue = Param(
        Params._dummy(),
        "maskValue",
        "Mask value to use for hash indexing",
        typeConverter=TypeConverters.toString,
    )

    def setNumBins(self, value: int) -> "HashIndexParams":
        """
        Sets the numBins parameter.

        :param value: Integer value for the number of bins to use for hash indexing.
        :returns: Instance of class mixed in.
        """
        if value <= 0:
            raise ValueError("Number of bins must be greater than 0.")
        return self._set(numBins=value)

    def getNumBins(self) -> int:
        """
        Gets the numBins parameter.

        :returns: Integer value for the number of bins to use for hash indexing.
        """
        return self.getOrDefault(self.numBins)

    def setMaskValue(self, value: str) -> "HashIndexParams":
        """
        Sets the maskValue parameter.

        :param value: String value for the mask value to use for hash indexing.
        :returns: Instance of class mixed in.
        """
        return self._set(maskValue=value)

    def getMaskValue(self) -> str:
        """
        Gets the maskValue parameter.

        :returns: String value for the mask value to use for hash indexing.
        """
        return self.getOrDefault(self.maskValue)


class PadValueParams(Params):
    """
    Mixin class containing pad value parameters needed
    for ordinal array encoder transformers and array crop transformers.
    """

    padValue = Param(
        Params._dummy(),
        "padValue",
        "The value to be considered as padding.",
        typeConverter=TypeConverters.identity,
    )

    def setPadValue(self, value: Union[str, int, float]) -> "PadValueParams":
        """
        Sets the parameter pad value to the given value.
        :param value: pad value.
        :returns: Instance of class mixed in.
        """
        return self._set(padValue=value)

    def getPadValue(self) -> Union[str, int, float]:
        """
        Gets the pad value parameter.
        :returns: string pad value.
        """
        return self.getOrDefault(self.padValue)


class ListwiseStatisticsParams(Params):
    """
    Mixin class containing the parameters needed for Listwise Statistics transformers.
    """

    queryIdCol = Param(
        Params._dummy(),
        "queryIdCol",
        """Column name to aggregate summary statistics upon,
        such as 'search_id'.""",
        typeConverter=TypeConverters.toString,
    )
    topN = Param(
        Params._dummy(),
        "topN",
        "Limit to how far into the list to aggregate.",
        typeConverter=TypeConverters.toInt,
    )
    minFilterValue = Param(
        Params._dummy(),
        "minFilterValue",
        """Value which equal to or greater than will be aggregated
         upon, anything less will be removed - this is primarily to deal
         with padded features.""",
        typeConverter=TypeConverters.toInt,
    )
    sortOrder = Param(
        Params._dummy(),
        "sortOrder",
        """Either leave as blank for default ordering or 'asc'
        for ascending values or 'desc' for descending values.""",
        typeConverter=TypeConverters.toString,
    )

    def setInputCols(self, value: List[str]) -> "ListwiseStatisticsParams":
        """
        Overrides setting the input columns for the transformer.
        Throws an error if we do not have exactly two input columns.
        :returns: Class instance.
        """
        if len(value) != 2:
            raise ValueError(
                """
                Arg inputCols must contain exactly two columns.
                The first is the value column, on which to calculate the statistic.
                The second is the sort column, based on which to sort the items.
                If you don't need sorting, use setInputCol instead.
                """
            )
        if self.getTopN() is None:
            raise ValueError("topN must be specified when using a sort column.")
        return self._set(inputCols=value)

    def setQueryIdCol(self, value: str) -> "ListwiseStatisticsParams":
        """
        Sets the query id parameter.

        :param value: String for column name to aggregate upon.
        :returns: Instance of class mixed in.
        """
        return self._set(queryIdCol=value)

    def getQueryIdCol(self) -> str:
        """
        Gets the query id parameter.

        :returns:  String for column name to aggregate upon.
        """
        return self.getOrDefault(self.queryIdCol)

    def setTopN(self, value: int) -> "ListwiseStatisticsParams":
        """
        Sets the top N parameter.

        :param value: Filter to limit length of input feature. Should be ordered first.
        :returns: Instance of class mixed in.
        """
        return self._set(topN=value)

    def getTopN(self) -> int:
        """
        Gets the top N parameter.

        :returns:  Filter to limit length of input feature. Should be ordered first.
        """
        return self.getOrDefault(self.topN)

    def setMinFilterValue(self, value: int) -> "ListwiseStatisticsParams":
        """
        Sets the min filter value parameter.

        :param value: Minimum value to remove padded values. Defaults to 0.
        :returns: Instance of class mixed in.
        """
        return self._set(minFilterValue=value)

    def getMinFilterValue(self) -> int:
        """
        Gets the min filter value parameter.

        :returns: Minimum value to remove padded values - defaults to >= 0.
        """
        return self.getOrDefault(self.minFilterValue)

    def setDefaultValue(self, value: str) -> "ListwiseStatisticsParams":
        """
        Sets the default value parameter.

        :param value: Int defaults to 0, should a method only have
        padded data this value will be returned.
        :returns: Instance of class mixed in.
        """
        return self._set(defaultValue=value)

    def getDefaultValue(self) -> int:
        """
        Gets the default value parameter.

        :returns: Int defaults to 0, should a method only have
        padded data this value will be returned.
        """
        return self.getOrDefault(self.defaultValue)

    def setSortOrder(self, value: str) -> "ListwiseStatisticsParams":
        """
        Sets the sortOrder parameter to the given value.
        Must be one of 'asc' or 'desc'.

        :param value: String to set the stringOrderType parameter to.
        :returns: Instance of class mixed in.
        """
        possible_order_options = [
            "asc",
            "desc",
        ]
        if value not in possible_order_options:
            raise ValueError(
                f"sortOrderType must be one of {', '.join(possible_order_options)}"
            )
        return self._set(sortOrder=value)

    def getSortOrder(self) -> str:
        """
        Gets the sort order parameter.

        :returns: String to set the stringOrderType parameter to.
        """
        return self.getOrDefault(self.sortOrder)


class MaskValueParams(Params):
    """
    Mixin class containing maskValue parameter needed
    for standard scale and imputation layers.
    This parameter is used to ignore certain values in the scaling process.
    For imputation, the value is ignored by the estimator and imputed over at
    training and inference.
    """

    maskValue = Param(
        Params._dummy(),
        "maskValue",
        """
        Value to be used as a mask by the transformer.
        """,
        typeConverter=TypeConverters.toFloat,
    )

    def setMaskValue(self, value: float) -> "MaskValueParams":
        """
        Sets the maskValue parameter.
        :param value: Float value to use as the mask value.
        :returns: Instance of class mixed in.
        """
        return self._set(maskValue=value)

    def getMaskValue(self) -> float:
        """
        Gets the maskValue parameter.
        :returns: Float value of the mask value.
        """
        return self.getOrDefault(self.maskValue)


class ImputeMethodParams(Params):
    """
    Mixin class containing imputeParam parameter for imputation layer.
    This parameter is used to select the method to estimate the value
    that should be imputed over the mask.
    """

    imputeMethod = Param(
        Params._dummy(),
        "imputeMethod",
        """
        Method with which to compute the value that is imputed over the mask value.
        """,
        typeConverter=TypeConverters.toString,
    )

    def __init__(self) -> None:
        super().__init__()
        self.valid_impute_methods = ["mean", "median"]

    def setImputeMethod(self, value: str) -> "ImputeMethodParams":
        """
        Sets the imputeParam parameter for the imputation layer.
        """
        if value not in self.valid_impute_methods:
            raise ValueError(
                "Imputation method should be one of: [{}], but received {}".format(
                    ",".join(self.valid_impute_methods), value
                )
            )
        return self._set(imputeMethod=value)

    def getImputeMethod(self) -> str:
        """
        Gets the imputeParam parameter value.
        :returns: Str value of the method to use to compute the value to impute.
        """
        return self.getOrDefault(self.imputeMethod)


class NanFillValueParams(Params):
    nanFillValue = Param(
        Params._dummy(),
        "nanFillValue",
        """
        The value to fill Nan with.
        """,
        typeConverter=TypeConverters.toFloat,
    )

    def setNanFillValue(self, value: float) -> "NanFillValueParams":
        """
        Sets the nanFillValue parameter.
        :param value: Float value to use as the fill value.
        :returns: Instance of class mixed in.
        """
        if value is None:
            raise ValueError("nanFillValue cannot be None")
        return self._set(nanFillValue=value)

    def getNanFillValue(self) -> float:
        """
        Gets the nanFillValue parameter.
        :returns: Float value of the fill value.
        """
        return self.getOrDefault(self.nanFillValue)


class StandardScaleSkipZerosParams(Params):
    """
    Mixin class containing maskValue parameter needed for conditional standard scale
    layers. This parameter is used to ignore zeros when scaling.
    """

    skipZeros = Param(
        Params._dummy(),
        "skipZeros",
        """
        If True, during spark transform and keras inference, do not apply the
        scaling when the values to scale are equal to zero.
        """,
        typeConverter=TypeConverters.toBoolean,
    )

    epsilon = Param(
        Params._dummy(),
        "epsilon",
        """
        Epsilon value to use when checking the skipZeros condition.
        """,
        typeConverter=TypeConverters.toFloat,
    )

    def setSkipZeros(self, value: bool) -> "StandardScaleSkipZerosParams":
        """
        Sets the skipZeros parameter.

        :param value: Boolean value to use as the mask value.
        :returns: Instance of class mixed in.
        """
        return self._set(skipZeros=value)

    def getSkipZeros(self) -> bool:
        """
        Gets the skipZeros parameter.

        :returns: Boolean value of the mask value.
        """
        return self.getOrDefault(self.skipZeros)

    def setEpsilon(self, value: float) -> "StandardScaleSkipZerosParams":
        """
        Sets the epsilon parameter.

        :param value: Float value to use as the epsilon value.
        :returns: Instance of class mixed in.
        """
        return self._set(epsilon=value)

    def getEpsilon(self) -> float:
        """
        Gets the epsilon parameter.

        :returns: Float value of the epsilon value.
        """
        return self.getOrDefault(self.epsilon)


class StandardScaleParams(MaskValueParams):
    """
    Mixin class containing mean and standard deviation parameters needed
    for standard scaler layers.
    """

    mean = Param(
        Params._dummy(),
        "mean",
        "Mean of the feature values.",
        typeConverter=TypeConverters.toListFloat,
    )

    stddev = Param(
        Params._dummy(),
        "stddev",
        "Standard deviation of the feature values.",
        typeConverter=TypeConverters.toListFloat,
    )

    def setMean(self, value: List[float]) -> "StandardScaleParams":
        """
        Sets the parameter mean to the given Vector value.
        Saves the mean as a list of floats.

        :param value: List of mean values.
        :returns: Instance of class mixed in.
        """
        if None in set(value):
            ids = [i for i, x in enumerate(value) if x is None]
            raise ValueError("Got null Mean values at positions: ", ids)
        return self._set(mean=value)

    def getMean(self) -> List[float]:
        """
        Gets the mean parameter.

        :returns: List of float mean values.
        """
        return self.getOrDefault(self.mean)

    def setStddev(self, value: List[float]) -> "StandardScaleParams":
        """
        Sets the parameter stddev to the given Vector value.
        Saves the standard deviation as a list of floats.

        :param value: Vector of standard deviation values.
        :returns: Instance of class mixed in.
        """
        if None in set(value):
            ids = [i for i, x in enumerate(value) if x is None]
            raise ValueError("Got null Stddev values at positions: ", ids)
        return self._set(stddev=value)

    def getStddev(self) -> List[float]:
        """
        Gets the stddev parameter.

        :returns: List of float standard deviation values.
        """
        return self.getOrDefault(self.stddev)


class UnixTimestampParams(Params):
    """
    Mixin class for a unix timestamp
    """

    unit = Param(
        Params._dummy(),
        "unit",
        """Unit of the timestamp.
        Can be `milliseconds` or `seconds`. Can also use short-hand `ms` and `s`.
        Default is `s` (seconds)""",
        typeConverter=TypeConverters.toString,
    )

    def setUnit(self, value: str) -> "UnixTimestampParams":
        """
        Sets the unit parameter.
        :param value: unit.
        :returns: Instance of class mixed in.
        """
        allowed_units = ["milliseconds", "seconds", "ms", "s"]
        if value not in allowed_units:
            raise ValueError(f"Unit must be one of {allowed_units}")

        if value == "milliseconds":
            value = "ms"
        if value == "seconds":
            value = "s"

        return self._set(unit=value)

    def getUnit(self) -> str:
        """
        Gets the unit parameter.
        :returns: unit.
        """
        return self.getOrDefault(self.unit)


class DateTimeParams(Params):
    """
    Mixin class for a datetime transformation
    """

    includeTime = Param(
        Params._dummy(),
        "includeTime",
        """Whether to include the time in the output datetime.
        If False, only the date is included in the format yyyy-MM-dd. If True,
        the time is also included in the format yyyy-MM-dd HH:mm:ss.SSS""",
        typeConverter=TypeConverters.toBoolean,
    )

    def setIncludeTime(self, value: bool) -> "DateTimeParams":
        """
        Sets the includeTime parameter.
        :param value: includeTime.
        :returns: Instance of class mixed in.
        """
        return self._set(includeTime=value)

    def getIncludeTime(self) -> bool:
        """
        Gets the includeTime parameter.
        :returns: includeTime.
        """
        return self.getOrDefault(self.includeTime)


class AutoBroadcastParams(Params):
    """
    Mixin class for the auto broadcast parameter.
    """

    autoBroadcast = Param(
        Params._dummy(),
        "autoBroadcast",
        """Whether to enable auto broadcast for the layer.
        If `True`, will broadcast the input tensors to the biggest rank before
        concatenating. Defaults to `False`.
        """,
        typeConverter=TypeConverters.toBoolean,
    )

    def setAutoBroadcast(self, value: bool) -> "AutoBroadcastParams":
        """
        Sets the autoBroadcast parameter.
        :param value: autoBroadcast.
        :returns: Instance of class mixed in.
        """
        return self._set(autoBroadcast=value)

    def getAutoBroadcast(self) -> bool:
        """
        Gets the autoBroadcast parameter.
        :returns: autoBroadcast.
        """
        return self.getOrDefault(self.autoBroadcast)


class LatLonConstantParams(Params):
    """
    Mixin class containing lat and lon constant parameters.
    """

    latLonConstant = Param(
        Params._dummy(),
        "latLonConstant",
        """Constant lat & lon to use in haversine distance calculation if only one
        input column is provided.""",
        typeConverter=TypeConverters.toListFloat,
    )

    def setLatLonConstant(self, value: List[float]) -> "LatLonConstantParams":
        """
        Sets the latLonConstant parameter.
        :param value: List of float lat and lon values.
        :returns: Instance of class mixed in.
        """
        if len(value) != 2:
            raise ValueError("latLonConstant must be a list of two floats: [lat, lon]")
        elif self.isDefined("inputCols") and len(self.getInputCols()) != 2:
            raise ValueError(
                f"""In order to set latLonConstant, there must be exactly two
                input columns. Found {len(self.getInputCols())} columns."""
            )
        elif value[0] < -90.0 or value[0] > 90.0:
            raise ValueError("Latitude must be between -90 and 90")
        elif value[1] < -180.0 or value[1] > 180.0:
            raise ValueError("Longitude must be between -180 and 180")
        return self._set(latLonConstant=value)

    def getLatLonConstant(self) -> List[float]:
        """
        Gets the latLonConstant parameter.
        :returns: List of float value of lat and lon used in haversine distance
        and bearing angle calculation.
        """
        return self.getOrDefault(self.latLonConstant)


class DefaultIntValueParams(Params):
    """
    Mixin class containing default integer parameter.
    """

    defaultValue = Param(
        Params._dummy(),
        "defaultValue",
        """
        Default int value to use in the transformer.
        """,
        typeConverter=TypeConverters.toInt,
    )

    def setDefaultValue(self, value: int) -> "DefaultIntValueParams":
        """
        Sets the defaultValue parameter.
        :param value: Value to set the defaultValue parameter to.
        :returns: Instance of class mixed in.
        """
        return self._set(defaultValue=value)

    def getDefaultValue(self) -> int:
        """
        Gets the defaultValue parameter.
        :returns: defaultValue param value.
        """
        return self.getOrDefault(self.defaultValue)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .date_utils import (  # noqa: F401
    datetime_add_days,
    datetime_day,
    datetime_day_of_year,
    datetime_hour,
    datetime_is_weekend,
    datetime_millisecond,
    datetime_minute,
    datetime_month,
    datetime_second,
    datetime_to_unix_timestamp,
    datetime_total_days,
    datetime_total_milliseconds,
    datetime_total_seconds,
    datetime_weekday,
    datetime_year,
    unix_timestamp_to_datetime,
)
from .input_utils import (  # noqa: F401
    allow_single_or_multiple_tensor_input,
    enforce_multiple_tensor_input,
    enforce_single_tensor_input,
)
from .list_utils import get_top_n, listify_tensors  # noqa: F401
from .shape_utils import reshape_to_equal_rank  # noqa: F401
from .transform_utils import map_fn_w_axis  # noqa: F401

from .layer_utils import NormalizeLayer  # noqa: F401 # isort:skip



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ArraySubtractMinimumLayer(BaseLayer):
    """
    TensorFlow layer that computes the difference across an axis from the minimum
    non-paded element in the input tensor.

    It takes a tensor of numerical value and calculates the differences between
    each value and the minimum value in the tensor. The calculation preserves
    the pad value elements.

    The principal use case for this layer is to calculate the time difference
    from the first event to all events in a sequence, where the tensor is a array of
    timestamps.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        axis: int = -1,
        pad_value=None,
        **kwargs,
    ) -> None:
        """
        Initialises the ArraySubtractMinimum layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: The axis along which the differences are calculated.
        Defaults to -1.
        :param pad_value: The value to be considered as padding. Defaults to `None`.
        :returns: None
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.axis = axis
        self.pad_value = pad_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.uint8,
            tf.int8,
            tf.uint16,
            tf.int16,
            tf.int32,
            tf.int64,
            tf.uint32,
            tf.uint64,
        ]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the calculation of the differences on the input tensor.

        Example:
         input_tensor = tf.Tensor([
            [19, 18, 13, 11, 10, -1, -1, -1],
            [12, 2, 1, -1, -1, -1, -1, -1],
            ]
         )
         layer = ArraySubtractMinimumLayer()
         differences = layer(input_tensor)
         print(differences)
         Output: tf.Tensor([[
            [9, 8, 3, 1, 0, -1, -1, -1],
            [11, 1, 0, -1, -1, -1, -1, -1],
            ]
        )

        :param inputs: The input tensor.
        :returns: Tensor of differences from the minimum (non-padded) timestamp.
        """
        if self.pad_value is None:
            # If pad value is not defined, then the smallest value in the tensor is
            # considered as the first value and subtracted from all the values.
            first_value = tf.reduce_min(inputs, axis=self.axis)
            subtracted_val = tf.subtract(inputs, tf.expand_dims(first_value, self.axis))
            return subtracted_val

        # Otherwise, we find the smallest non padded value and subtract it from all
        # the values. Padded values are preserved.
        inputs, pad_tensor = self._force_cast_to_compatible_numeric_type(
            inputs, self.pad_value
        )
        first_non_pad_value = tf.reduce_min(
            tf.where(tf.equal(inputs, pad_tensor), inputs.dtype.max, inputs),
            axis=self.axis,
        )
        subtracted_val = tf.subtract(
            inputs, tf.expand_dims(first_non_pad_value, self.axis)
        )
        return tf.where(tf.equal(inputs, pad_tensor), inputs, subtracted_val)

    def get_config(self):
        """
        Returns the configuration of the layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer
        """
        config = super().get_config()
        config.update(
            {
                "pad_value": self.pad_value,
                "axis": self.axis,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class LambdaFunctionLayer(BaseLayer, tf.keras.layers.Lambda):
    """
    Performs the lambda function operation on a given input tensor

    WARNING: This layer relies on a `tf.keras.layers.Lambda` layer which have
    (de)serialization limitations!

    `Lambda` layers are saved by serializing the Python bytecode, which is fundamentally
    non-portable. They should only be loaded in the same environment where
    they were saved.
    """

    def __init__(
        self,
        function: Callable[[Union[Tensor, List[Tensor]]], Union[Tensor, List[Tensor]]],
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initializes the LambdaFunction layer

        :param function: The lambda function to apply to the input tensor(s).
        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name,
            input_dtype=input_dtype,
            output_dtype=output_dtype,
            function=function,
            **kwargs,
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @allow_single_or_multiple_tensor_input
    def _call(
        self, inputs: Union[Tensor, Iterable[Tensor]], **kwargs
    ) -> Union[Tensor, Iterable[Tensor]]:
        """
        Transforms the input tensor(s) by applying the lambda function.

        Decorated with `@allow_single_or_multiple_tensor_input` to ensure that the input
        is either a single tensor or an iterable of tensors. Returns this result as a
        list of tensors for easier use here.

        :param inputs: Tensor(s) to apply the lambda function to.
        :returns: The transformed tensor(s).
        """
        if len(inputs) == 1:
            return tf.keras.layers.Lambda.call(self, inputs[0], **kwargs)
        return tf.keras.layers.Lambda.call(self, inputs, **kwargs)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the LambdaFunction layer.
        Used for saving and loading from a model.
        Calls the parent class's get_config method which deals with serialising the
        function.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import eq, ge, gt, le, lt, ne
from typing import Any, Callable


def get_condition_operator(cond_op_string: str) -> Callable[[Any, Any], Any]:
    """
    Translates a string condition operator to a function operator.

    :returns: Function operator.
    """
    allowed_cond_ops = {
        "eq": eq,
        "neq": ne,
        "lt": lt,
        "leq": le,
        "gt": gt,
        "geq": ge,
    }
    try:
        return allowed_cond_ops[cond_op_string]
    except KeyError:
        raise ValueError(
            f"""Unknown condition operator: {cond_op_string}.
            Allowed condition operators are: {allowed_cond_ops.keys()}"""
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Union

import keras_tuner as kt
import tensorflow as tf
from sklearn.pipeline import Pipeline

from kamae.graph import PipelineGraph


class KamaeSklearnPipeline(Pipeline):
    """
    KamaeSklearnPipeline is a subclass of sklearn.pipeline.Pipeline that is used to
    chain together BaseTransformers. It maintains the same functionality
    as sklearn.pipeline.Pipeline e.g. serialisation.
    """

    def __init__(self, steps, *, memory=None, verbose=False) -> None:
        """
        Initializes a KamaeSklearnPipeline object.

        :param steps: List of tuples containing the name and LayerTransformer
        :param memory: str or object with the joblib.Memory interface, default=None
        Used to cache the fitted transformers of the pipeline. The last step
        will never be cached, even if it is a transformer. By default, no
        caching is performed. If a string is given, it is the path to the
        caching directory. Enabling caching triggers a clone of the transformers
        before fitting. Therefore, the transformer instance given to the
        pipeline cannot be inspected directly. Use the attribute ``named_steps``
        or ``steps`` to inspect estimators within the pipeline. Caching the
        transformers is advantageous when fitting is time consuming.
        :param verbose: If True, the time elapsed while fitting each step
        will be printed as it is completed.
        """
        super().__init__(steps, memory=memory, verbose=verbose)

    def get_all_tf_layers(self) -> List[tf.keras.layers.Layer]:
        """
        Gets a list of all tensorflow layers in the pipeline model.

        :returns: List of tensorflow layers within the pipeline model.
        """
        return [step[1].get_tf_layer() for step in self.steps]

    def build_keras_model(
        self,
        tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]],
        output_names: Optional[List[str]] = None,
    ) -> tf.keras.Model:
        """
        Builds a keras model from the pipeline using the PipelineGraph
        helper class.

        :param tf_input_schema: List of dictionaries containing the input schema for
        the model. Specifically the name, shape and dtype of each input.
        These will be passed as is to the Keras Input layer.
        :param output_names: Optional list of output names for the Keras model. If
        provided, only the outputs specified are used as model outputs.
        :returns: Keras model.
        """
        stage_dict = {
            step[1].layer_name: step[1].construct_layer_info() for step in self.steps
        }
        pipeline_graph = PipelineGraph(stage_dict=stage_dict)
        return pipeline_graph.build_keras_model(
            tf_input_schema=tf_input_schema, output_names=output_names
        )

    def get_keras_tuner_model_builder(
        self,
        tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]],
        hp_dict: Dict[str, List[Dict[str, Any]]],
        output_names: Optional[List[str]] = None,
    ) -> Callable[[kt.HyperParameters], tf.keras.Model]:
        """
        Builds a keras tuner model builder (function) from the pipeline model
        using the PipelineGraph helper class.

        :param tf_input_schema: List of dictionaries containing the input schema for
        the model. Specifically the name, shape and dtype of each input.
        These will be passed as is to the Keras Input layer.
        :param hp_dict: Dictionary containing the hyperparameters for the model.
        :param output_names: Optional list of output names for the Keras model. If
        provided, only the outputs specified are used as model outputs.
        :returns: Keras tuner model builder (function).
        """
        stage_dict = {
            step[1].layer_name: step[1].construct_layer_info() for step in self.steps
        }
        pipeline_graph = PipelineGraph(stage_dict=stage_dict)
        return pipeline_graph.get_keras_tuner_model_builder(
            tf_input_schema=tf_input_schema, hp_dict=hp_dict, output_names=output_names
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import tensorflow as tf


def add_missing_time_components_to_datetime_tensor(
    datetime_tensor: tf.Tensor, max_len: Optional[int] = None
) -> tf.Tensor:
    """
    Adds missing time components to a date string tensor.
    If the time components are missing, they will be added as zeros.

    :param datetime_tensor: date string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format. Can be truncated, and missing time
    components will be added as zeros.
    :param max_len: Maximum length to append time to if the time is missing. Used to
    avoid unnecessary computation. E.g. if we only need hour, then don't add
    milliseconds. Default is None.
    :returns: Date string tensor with missing time components added as zeros.
    """
    if max_len is not None and max_len < 10:
        raise ValueError(
            """max_len must be at least 10, as this is the minimum length
            of a date string."""
        )
    # Add missing time components, these are at 10, 13, 16 and 19 characters
    # For hours, minutes, seconds and milliseconds respectively
    str_lens = [10, 13, 16, 19]
    str_suffixes = [" 00:00:00.000", ":00:00.000", ":00.000", ".000"]
    # Filter out the suffixes that are longer than the max_len. This allows us to not
    # add time components if we don't need them.
    str_loop = (
        filter(lambda x: x[0] <= max_len, zip(str_lens, str_suffixes))
        if max_len is not None
        else zip(str_lens, str_suffixes)
    )
    for str_len, str_suffix in str_loop:
        dynamic_str_len = tf.strings.length(datetime_tensor)
        datetime_tensor = tf.where(
            dynamic_str_len == str_len,
            tf.strings.join([datetime_tensor, str_suffix], ""),
            datetime_tensor,
        )
    return datetime_tensor


def datetime_days_to_month(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Helper function for some datetime functions.
    Gets the number of days to the month of the given datetime tensor.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Number of days to month, stored as tf.int64.
    """
    # 30 days have September...
    days_in_month = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    # Extract date parts
    year = datetime_year(datetime_tensor)
    month = datetime_month(datetime_tensor)
    days_to_month = tf.reduce_sum(
        tf.stack(
            [
                tf.where(month > idx + 1, 1, 0) * n_days
                for idx, n_days in enumerate(days_in_month)
            ],
            axis=-1,
        ),
        -1,
    ) + (
        tf.where(month > 2, 1, 0)
        * tf.where((year % 4 == 0) & ((year % 100 != 0) | (year % 400 == 0)), 1, 0)
    )

    days_to_month = tf.cast(days_to_month, tf.int64)

    return days_to_month


def datetime_year(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a year tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Year tensor, stored as tf.int64.
    """
    year = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 0, 4), out_type=tf.int64
    )
    return year


def datetime_month(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a month tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Month tensor, stored as tf.int64.
    """
    month = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 5, 2), out_type=tf.int64
    )
    return month


def datetime_day(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a day tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Day tensor, stored as tf.int64.
    """
    day = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 8, 2), out_type=tf.int64
    )
    return day


def datetime_hour(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into an hour tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Hour tensor, stored as tf.int64.
    """
    datetime_tensor = add_missing_time_components_to_datetime_tensor(
        datetime_tensor, max_len=13
    )
    hour = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 11, 2), out_type=tf.int64
    )
    return hour


def datetime_minute(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a minute tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Minute tensor, stored as tf.int64.
    """
    datetime_tensor = add_missing_time_components_to_datetime_tensor(
        datetime_tensor, max_len=16
    )
    minute = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 14, 2), out_type=tf.int64
    )
    return minute


def datetime_second(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a second tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Second tensor, stored as tf.int64.
    """
    datetime_tensor = add_missing_time_components_to_datetime_tensor(
        datetime_tensor, max_len=19
    )
    second = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 17, 2), out_type=tf.int64
    )
    return second


def datetime_millisecond(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a millisecond tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Millisecond tensor, stored as tf.int64.
    """
    datetime_tensor = add_missing_time_components_to_datetime_tensor(datetime_tensor)
    millisecond = tf.strings.to_number(
        tf.strings.substr(datetime_tensor, 20, 3), out_type=tf.int64
    )
    return millisecond


def datetime_total_days(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a total days tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Total days tensor, stored as tf.int64.
    """
    year = datetime_year(datetime_tensor)
    day = datetime_day(datetime_tensor)
    first_century_year_post_1970 = tf.constant([2000], dtype=tf.int64)
    num_standard_days = (year - 1970) * 365
    # Compute the number of leap years to know if we need to add extra days.
    # We only consider year - 1, since if we are currently in a leap year, this will
    # be catered for in days_to_month.
    num_standard_leap_years = ((year - 1) - 1972) // 4
    num_century_years = tf.where(
        year > first_century_year_post_1970,
        ((year - 1) - first_century_year_post_1970) // 100,
        0,
    )
    num_century_leap_years = tf.where(
        year > first_century_year_post_1970,
        ((year - 1) - first_century_year_post_1970) // 400,
        0,
    )
    # Subtract all century years and add all century leap years.
    num_leap_years = (
        num_standard_leap_years - num_century_years + num_century_leap_years
    )
    # Days to year is the number of standard days across all the years plus the number
    # of leap years (as each leap year adds exactly 1 day)
    days_to_year = num_standard_days + num_leap_years
    days_to_month = datetime_days_to_month(datetime_tensor)
    # Add all the days together
    total_days = days_to_year + days_to_month + day

    return total_days


def datetime_total_seconds(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a total seconds tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Total seconds tensor, stored as tf.int64.
    """
    # Extract date parts
    total_days = tf.cast(datetime_total_days(datetime_tensor), dtype=tf.float64)
    hour = tf.cast(datetime_hour(datetime_tensor), dtype=tf.float64)
    minute = tf.cast(datetime_minute(datetime_tensor), dtype=tf.float64)
    second = tf.cast(datetime_second(datetime_tensor), dtype=tf.float64)
    milliseconds = tf.cast(datetime_millisecond(datetime_tensor), dtype=tf.float64)
    # Add all the seconds together
    total_seconds = (
        (total_days * 24 * 60 * 60)
        + (hour * 60 * 60)
        + (minute * 60)
        + second
        + (milliseconds / tf.constant(1000.0, dtype=tf.float64))
    )
    return total_seconds


def datetime_total_milliseconds(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a total milliseconds tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Total milliseconds tensor, stored as tf.int64.
    """
    # Extract date parts
    total_days = datetime_total_days(datetime_tensor)
    hour = datetime_hour(datetime_tensor)
    minute = datetime_minute(datetime_tensor)
    second = datetime_second(datetime_tensor)
    millisecond = datetime_millisecond(datetime_tensor)
    # Add all the milliseconds together
    total_milliseconds = (
        (total_days * 24 * 60 * 60 * 1000)
        + (hour * 60 * 60 * 1000)
        + (minute * 60 * 1000)
        + (second * 1000)
        + millisecond
    )
    return total_milliseconds


def datetime_weekday(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a weekday tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Weekday tensor, stored as tf.int64.
    """
    total_days = datetime_total_days(datetime_tensor)
    # Compute the weekday
    week_day = (total_days - 4) % 7 + 1
    return week_day


def datetime_is_weekend(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a weekend tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Weekend tensor, stored as tf.int64.
    """
    week_day = datetime_weekday(datetime_tensor)
    # Compute the weekend
    is_weekend = tf.cast(tf.where(week_day > 5, 1, 0), tf.int64)
    return is_weekend


def datetime_day_of_year(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Utility function to parse a date(time) tensor into a day of year tensor.
    Uses native tf functions only to avoid serialization issues.

    :param: datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown, and you will get a nonsense output.

    :returns: Day of year tensor, stored as tf.int64.
    """
    day = datetime_day(datetime_tensor)
    days_to_month = datetime_days_to_month(datetime_tensor)
    # Add all the days together
    day_of_year = days_to_month + day

    return day_of_year


def datetime_add_days(
    datetime_tensor: tf.Tensor, num_days: tf.Tensor, include_time: bool = True
) -> tf.Tensor:
    """
    Adds a number of days to a date(time) string tensor.

    :param datetime_tensor: date(time) string tensor.
    Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.
    :param num_days: Number of days to add.
    :param include_time: Whether to include the time in the output. If True, the output
    will be in yyyy-MM-dd HH:mm:ss.SSS format. If False, the output will be in
    yyyy-MM-dd format. Default is True.
    :returns: Date(time) string tensor with num_days added.
    """
    total_seconds = datetime_total_seconds(datetime_tensor)
    num_days_seconds = num_days * tf.constant(24 * 60 * 60, dtype=num_days.dtype)
    total_seconds += num_days_seconds
    return unix_timestamp_to_datetime(
        tf.cast(total_seconds, dtype=tf.float64), include_time=include_time
    )


def unix_timestamp_to_datetime(
    timestamp_tensor: tf.Tensor, include_time: bool = True
) -> tf.Tensor:
    """
    Converts a timestamp tensor (seconds since Unix Epoch) into a datetime string
    tensor. If include_time is False, the output will be in yyyy-MM-dd, if include_time
    is True, the output will be in yyyy-MM-dd HH:mm:ss.SSS format.

    :param timestamp_tensor: the timestamp tensor to convert.
    Timestamps must be in seconds since unix epoch.
    :param include_time: Whether to include the time in the output. If True, the output
    will be in yyyy-MM-dd HH:mm:ss.SSS format. If False, the output will be in
    yyyy-MM-dd format. Default is True.
    :returns: Datetime string tensor in either yyyy-MM-dd or yyyy-MM-dd HH:mm:ss.SSS
    format.
    """

    # Days, hours, minutes and seconds since Unix Epoch
    seconds_in_one_minute = tf.constant(60.0, dtype=tf.float64)
    seconds_in_one_hour = tf.math.multiply(seconds_in_one_minute, 60.0)
    seconds_in_one_day = tf.math.multiply(seconds_in_one_hour, 24.0)
    total_days = tf.math.floordiv(timestamp_tensor, seconds_in_one_day)

    # Initialise the remainder days variable
    remainder_days = total_days
    days_in_4_years = tf.constant(1461.0, dtype=tf.float64)
    year = tf.add(
        tf.constant(1970.0, dtype=tf.float64),
        tf.multiply(
            tf.math.floordiv(remainder_days, days_in_4_years),
            tf.constant(4.0, dtype=tf.float64),
        ),
    )
    remainder_days = tf.math.mod(remainder_days, days_in_4_years)

    # Let k = the number of 4 year chunks since 1970
    # We count from 1970 + 4k, so every 3rd year is a leap year
    # (e.g. 1970 + 4k, 1971 + 4k, ^^1972 + 4^^)
    # We don't need to count the last year as the remainder will get
    # carried on to the next loop where the month is computed
    # TODO: Is there a better abstraction instead of for loops?
    #  These are O(1) operations, but feel clunky and also not very clear
    year_days = [
        tf.constant(365.0, dtype=tf.float64),
        tf.constant(365.0, dtype=tf.float64),
        tf.constant(366.0, dtype=tf.float64),
    ]
    for d in year_days:
        year_passed = tf.where(
            remainder_days >= d,
            tf.constant(1.0, dtype=tf.float64),
            tf.constant(0.0, dtype=tf.float64),
        )
        year += year_passed
        remainder_days -= year_passed * d

    # The full days in year that have been realised
    full_days_in_year = remainder_days

    # Initialise month loop variables
    # Days in the month (we treat leap years in the loop)
    month_days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    months_to_month = tf.zeros_like(total_days)
    remainder_days = full_days_in_year

    # First loop starts from December and works backwards
    for idx, _ in enumerate(month_days):
        n_months = 12 - idx

        cumulative_days_to_month = (
            # Leap year treatment (if we are in a leap year)
            # A leap year is one that is divisible by 4, unless it is divisible by 100
            # but not divisible by 400
            (
                tf.where(
                    (year % 4 == 0) & ((year % 100 != 0) | (year % 400 == 0)),
                    tf.constant(1.0, dtype=tf.float64),
                    tf.constant(0.0, dtype=tf.float64),
                )
                * tf.where(
                    n_months >= 2,
                    tf.constant(1.0, dtype=tf.float64),
                    tf.constant(0.0, dtype=tf.float64),
                )
            )
            # Cumulative days in a normal year
            + sum(month_days[:n_months])
        )

        # Elements will be zero unless ALL cumulative_days_to_month have been realised,
        # in which case the element will be 1
        month_has_been_realised = remainder_days // cumulative_days_to_month
        remainder_days -= month_has_been_realised * cumulative_days_to_month
        months_to_month += n_months * month_has_been_realised

    # The month we are in hasn't been realised fully, but we are in it (so +1)
    month = months_to_month + 1
    # The day we are in has not been realised fully, but we are in it (so +1)
    day = remainder_days + 1

    year_str = tf.strings.as_string(tf.cast(year, dtype=tf.int64))
    month_str = tf.strings.as_string(tf.cast(month, dtype=tf.int64), width=2, fill="0")
    day_str = tf.strings.as_string(tf.cast(day, dtype=tf.int64), width=2, fill="0")
    date = tf.strings.join([year_str, month_str, day_str], "-")

    if include_time:
        leftover_seconds = timestamp_tensor - tf.math.multiply(
            total_days, seconds_in_one_day
        )
        total_hours = tf.math.floordiv(leftover_seconds, seconds_in_one_hour)
        leftover_seconds -= tf.math.multiply(total_hours, seconds_in_one_hour)

        total_mins = tf.math.floordiv(leftover_seconds, seconds_in_one_minute)
        leftover_seconds -= tf.math.multiply(total_mins, seconds_in_one_minute)
        total_seconds = tf.math.floor(leftover_seconds)
        total_milliseconds = leftover_seconds - total_seconds

        hours_str = tf.strings.as_string(
            tf.cast(total_hours, dtype=tf.int64), width=2, fill="0"
        )
        minutes_str = tf.strings.as_string(
            tf.cast(total_mins, dtype=tf.int64), width=2, fill="0"
        )
        seconds_str = tf.strings.as_string(
            tf.cast(total_seconds, dtype=tf.int64), width=2, fill="0"
        )
        milliseconds_str = tf.strings.as_string(
            # We need to round the milliseconds to fix them to 3 decimal places
            tf.cast(tf.math.round(total_milliseconds * 1000.0), tf.int64),
            width=3,
            fill="0",
        )

        time = tf.strings.join(
            [
                tf.strings.join([hours_str, minutes_str, seconds_str], ":"),
                milliseconds_str,
            ],
            ".",
        )
        datetime = tf.strings.join([date, time], " ")
        return datetime

    return date


def datetime_to_unix_timestamp(datetime_tensor: tf.Tensor) -> tf.Tensor:
    """
    Converts a date string tensor into a timestamp tensor (seconds since Unix Epoch).

    :param datetime_tensor: the date tensor to convert.
        Must be in yyyy-MM-dd (HH:mm:ss.SSS) format.
    :returns: Timestamp tensor in seconds since Unix Epoch
    """
    return datetime_total_seconds(datetime_tensor)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import (
    allow_single_or_multiple_tensor_input,
    datetime_add_days,
)

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class DateAddLayer(BaseLayer):
    """
    Adds or subtracts a number of days from a date(time) string.

    WARNING: This layer destroys the time component of the date column.
    """

    def __init__(
        self,
        name=None,
        input_dtype: str = None,
        output_dtype: str = None,
        num_days: int = None,
        **kwargs,
    ) -> None:
        """
        Initialises an instance of the DateAddLayer.

        :param num_days: Number of days to add or subtract.
        :param name: Name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if num_days is not None and not isinstance(num_days, int):
            raise ValueError(
                f"Expected `num_days` to be an integer, but got {num_days}."
            )
        if num_days is None and input_dtype is not None:
            raise ValueError(
                """When `num_days` is not set, the layer expects two inputs of different
                dtypes. Therefore input auto-casting via `input_dtype` is not supported.
                """
            )
        self.num_days = num_days

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string, tf.int8, tf.int16, tf.int32, tf.int64]

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Adds or subtracts a number of days from a date(time) string.
        """
        if inputs[0].dtype != tf.string:
            raise ValueError(
                f"Expected input dtype to be tf.string, but got {inputs[0].dtype}."
            )
        if self.num_days is not None:
            if len(inputs) > 1:
                raise ValueError(
                    "When `num_days` is set, the input should be a single tensor."
                )
            return datetime_add_days(
                inputs[0],
                tf.constant(self.num_days, dtype=tf.float64),
                include_time=False,
            )
        else:
            if len(inputs) != 2:
                raise ValueError(
                    "When `num_days` is not set, the input should be two tensors."
                )
            if not inputs[1].dtype.is_integer:
                raise ValueError(
                    f"""Expected second input dtype to be integer, but got
                    {inputs[1].dtype}."""
                )
            return datetime_add_days(
                inputs[0],
                # Casting is necessary since all datetime ops are in float64
                # Furthermore, due to the input dtypes being different (e.g. first input
                # must be tf.string, second input must be integer), we cast to
                # potentially undo the auto-casting done by specifying input_dtype.
                self._cast(inputs[1], cast_dtype="float64"),
                include_time=False,
            )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the DateAdd layer.
        Used for saving and loading from a model.

        Specifically adds the `num_days` to the config dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"num_days": self.num_days})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf
from tensorflow.math import atan2, cos, mod, sin

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class BearingAngleLayer(BaseLayer):
    """
    Computes the Bearing angle operation on a given input tensor.
    If lat_lon_constant is not set, inputs must be a list of 4 tensors,
    in the order of lat1, lon1, lat2, lon2.
    If lat_lon_constant is set, inputs must be a tensor of 2 tensors,
    in the order of lat1, lon1.

    We DO NOT check if the lat/lon values are out of bounds.
    For lat, this is [-90, 90] and for lon, this is [-180, 180].
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        lat_lon_constant: List[float] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the BearingAngleLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param lat_lon_constant: The lat/lons to use in the bearing angle
        calculation. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if lat_lon_constant is not None and len(lat_lon_constant) != 2:
            raise ValueError("If set, lat_lon_constant must be a list of 2 floats")
        self.lat_lon_constant = lat_lon_constant

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.bfloat16, tf.float16, tf.float32, tf.float64]

    @staticmethod
    def get_radians(degrees: Tensor) -> Tensor:
        """
        Converts degrees tensor to radians. We need to cast to float64 otherwise
        pi / 180 will lose precision.

        :param degrees: Tensor of degrees.
        :returns: Tensor of radians.
        """
        return tf.cast(degrees, dtype=tf.float64) * tf.constant(
            math.pi / 180, dtype=tf.float64
        )

    @staticmethod
    def get_degrees(radians: Tensor) -> Tensor:
        """
        Converts radians tensor to degrees.

        :param radians: Tensor of degrees.
        :returns: Tensor of degrees.
        """
        return tf.cast(radians, dtype=tf.float64) * tf.constant(
            180 / math.pi, dtype=tf.float64
        )

    def compute_bearing_angle(
        self, lat1: Tensor, lon1: Tensor, lat2: Tensor, lon2: Tensor
    ) -> Tensor:
        """
        Computes the bearing angle between two lat/lon pairs.

        :param lat1: Tensor of latitudes of the first point.
        :param lon1: Tensor of longitudes of the first point.
        :param lat2: Tensor of latitudes of the second point.
        :param lon2: Tensor of longitudes of the second point.
        :returns: Tensor of bearing angles.
        """
        lat1_radians = self.get_radians(lat1)
        lon1_radians = self.get_radians(lon1)
        lat2_radians = self.get_radians(lat2)
        lon2_radians = self.get_radians(lon2)

        lon_difference = lon2_radians - lon1_radians
        # Bearing formula calculation
        y = sin(lon_difference) * cos(lat2_radians)

        x = cos(lat1_radians) * sin(lat2_radians)
        x -= sin(lat1_radians) * cos(lat2_radians) * cos(lon_difference)

        # Calculate bearing in degrees
        bearing = atan2(y, x)
        bearing_deg = mod(self.get_degrees(bearing) + 360, 360)
        return bearing_deg

    @enforce_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs) -> Tensor:
        """
        Computes the bearing angle between two lat/lon pairs.

        Decorated with @enforce_multiple_tensor_input to ensure that the input
        is an iterable of tensors. Raises an error if a single tensor is passed.

        After decoration, we check the length of the inputs to ensure we have the right
        number of lat/lon tensors.

        :param inputs: Iterable of tensors.
        :returns: Tensor of bearing angles.
        """
        if self.lat_lon_constant is not None:
            if not isinstance(inputs, list) or len(inputs) != 2:
                raise ValueError(
                    """If lat_lon_constant is set,
                inputs must be a list of 2 tensors"""
                )
            return self.compute_bearing_angle(
                inputs[0],
                inputs[1],
                tf.constant(self.lat_lon_constant[0]),
                tf.constant(self.lat_lon_constant[1]),
            )
        else:
            if not isinstance(inputs, list) or len(inputs) != 4:
                raise ValueError(
                    """If lat_lon_constant is not set,
                inputs must be a list of 4 tensors"""
                )
            return self.compute_bearing_angle(
                inputs[0],
                inputs[1],
                inputs[2],
                inputs[3],
            )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the Bearing Angle layer.
        Used for saving and loading from a model.

        Specifically, we add the `lat_lon_constant` and `unit` to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"lat_lon_constant": self.lat_lon_constant})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from itertools import chain
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import DefaultIntValueParams, SingleInputSingleOutputParams
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import DateParseLayer


class DateParseParams(DefaultIntValueParams):
    """
    Mixin class for a date part.
    """

    datePart = Param(
        Params._dummy(),
        "datePart",
        "Date part to extract from date",
        typeConverter=TypeConverters.toString,
    )

    def getDatePart(self) -> str:
        """
        Gets the value of the datePart parameter.

        :returns: Date part to extract from date.
        """
        return self.getOrDefault(self.datePart)

    def setDatePart(self, value: str) -> "DateParseParams":
        """
        Sets the value of the datePart parameter.

        :param value: Date part to extract from date.
        :returns: Class instance.
        """

        allowed_date_parts = {
            "Year",
            "DayOfYear",
            "MonthOfYear",
            "DayOfMonth",
            "DayOfWeek",
            "Hour",
            "Minute",
            "Second",
            "Millisecond",
        }

        if value not in allowed_date_parts:
            raise ValueError(
                f"Invalid date part: {value}. Must be one of {allowed_date_parts}"
            )

        return self._set(datePart=value)


class DateParseTransformer(
    BaseTransformer, SingleInputSingleOutputParams, DateParseParams
):
    """
    Date parse transform layer.
    This layer parses a date(time) column into a specified date part.
    We require the date format to be yyyy-MM-dd (HH:mm:ss.SSS).

    Date parts can be one of the following:
    - `DayOfWeek` - day of week (Monday = 1, Sunday = 7)
    - `DayOfMonth` - day of month
    - `DayOfYear` - day of year e.g. (2021-01-01 = 1, 2021-12-31 = 365)
    - `MonthOfYear` - month of year
    - `Year` - year
    - `Hour` - hour e.g. (2021-01-01 00:00:00 = 0, 2021-01-01 23:59:59 = 23)
    - `Minute` - minute e.g. (2021-01-01 00:00:00 = 0, 2021-01-01 00:59:00 = 59)
    - `Second` - second e.g. (2021-01-01 00:00:00 = 0, 2021-01-01 00:00:59 = 59)
    - `Millisecond` - millisecond (2021-01-01 00:00:00.357 = 357)

    In the case a timestamp is not provided, all hour, minutes, seconds and milliseconds
    fields will be returned as 0.
    """

    @keyword_only
    def __init__(
        self,
        datePart: Optional[str] = None,
        defaultValue: Optional[int] = None,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initialises the date parse transform layer.

        :param datePart: Date part to extract from date.
        :param defaultValue: Default value to use when the date is the empty string.
        Empty strings can be used when the date is not available.
        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Layer name. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(defaultValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input date column into the specified date part.

        Utilises date format, which itself uses the following:
        https://spark.apache.org/docs/latest/sql-ref-datetime-pattern.html

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def date_parse(x: Column) -> Column:
            if self.getDefaultValue() is not None:
                return F.when(x == F.lit(""), F.lit(self.getDefaultValue())).otherwise(
                    self._parse_date(x)
                )
            return self._parse_date(x)

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: date_parse(x),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def _parse_date(self, column: Column) -> Column:
        """
        Private function to parse the date column.
        :param column: Input column.
        :returns: Parsed datepart column.
        """

        date_part_to_format_pattern = {
            "Year": "y",
            "DayOfYear": "D",
            "MonthOfYear": "M",
            "DayOfMonth": "d",
            "DayOfWeek": "E",
            "Hour": "H",
            "Minute": "m",
            "Second": "s",
            "Millisecond": "SSS",
        }

        formatted_date = F.date_format(
            column, date_part_to_format_pattern[self.getDatePart()]
        )

        if self.getDatePart() == "DayOfWeek":
            day_of_week_mapping = {
                "Mon": 1,
                "Tue": 2,
                "Wed": 3,
                "Thu": 4,
                "Fri": 5,
                "Sat": 6,
                "Sun": 7,
            }

            day_of_week_spark_mapping = F.create_map(
                [F.lit(x) for x in chain(*day_of_week_mapping.items())]
            )

            return day_of_week_spark_mapping[formatted_date]

        return formatted_date.cast("int")

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer.

        :returns: DateParseLayer Tensorflow layer.
        """

        if not self.isDefined("datePart"):
            raise ValueError("Date part must be set.")
        date_part = self.getDatePart()

        return DateParseLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            date_part=date_part,
            default_value=self.getDefaultValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringAffixLayer

from .base import BaseTransformer


class StringAffixParams(Params):
    """
    Mixin class containing parameters needed for string affixing.
    transforms.
    """

    prefix = Param(
        Params._dummy(),
        "prefix",
        "Value to use as a prefix when joining the strings.",
        typeConverter=TypeConverters.toString,
    )
    suffix = Param(
        Params._dummy(),
        "suffix",
        "Value to use as a suffix when joining the strings.",
        typeConverter=TypeConverters.toString,
    )

    def setPrefix(self, value: str) -> "StringAffixParams":
        """
        Sets the prefix parameter.

        :param value: String value to use as a prefix when joining the strings.
        :returns: Instance of class mixed in.
        """
        return self._set(prefix=value)

    def getPrefix(self) -> str:
        """
        Gets the prefix parameter.

        :returns: String value to use as a prefix when joining the strings.
        """
        return self.getOrDefault(self.prefix)

    def setSuffix(self, value: str) -> "StringAffixParams":
        """
        Sets the suffix parameter.

        :param value: String value to use as a suffix when joining the strings.
        :returns: Instance of class mixed in.
        """
        return self._set(suffix=value)

    def getSuffix(self) -> str:
        """
        Gets the suffix parameter.

        :returns: String value to use as a suffix when joining the strings.
        """
        return self.getOrDefault(self.suffix)


class StringAffixTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    StringAffixParams,
):
    """
    String Affix Spark Transformer for use in Spark pipelines.
    This transformer takes in a column and pre- and su- fixes it.
    Input columns must be of type string.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> None:
        """
        Initializes the string affix transformer.
        :param inputCol: column to combine with prefix or suffix. Must be type string.
        :param outputCol: column to output the affixed string to.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param prefix: String to use as a prefix when joining the strings.
        :param suffix: String to use as a suffix when joining the strings.
        """
        super().__init__()
        self._setDefault(prefix=None, suffix=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _validate_params(self) -> None:
        """
        Validates the parameters passed to the transformer.
        """
        prefix = self.getPrefix()
        suffix = self.getSuffix()
        if (prefix is None or prefix == "") and (suffix is None or suffix == ""):
            raise ValueError(
                "Either prefix or suffix must be set. Otherwise nothing to affix."
            )

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        where the value is origin column combined with prefix and or suffix.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        self._validate_params()

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def add_prefix_suffix(
            column: Column, prefix: Optional[str] = None, suffix: Optional[str] = None
        ):
            if prefix is not None and prefix != "":
                column = F.concat(F.lit(prefix), column)
            if suffix is not None and suffix != "":
                column = F.concat(column, F.lit(suffix))
            return column

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: add_prefix_suffix(x, self.getPrefix(), self.getSuffix()),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the string affix transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs prefixing and suffixing.
        """
        return StringAffixLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            prefix=self.getPrefix(),
            suffix=self.getSuffix(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringCaseLayer

from .base import BaseTransformer


class StringCaseParams(Params):
    """
    Mixin class containing stringCaseType parameter needed for string casing transforms.
    """

    stringCaseType = Param(
        Params._dummy(),
        "stringCaseType",
        "How to change the case of the string.",
        typeConverter=TypeConverters.toString,
    )

    def setStringCaseType(self, value: str) -> "StringCaseParams":
        """
        Sets the stringCaseType parameter to the given value.
        Must be one of:
        - 'upper'
        - 'lower'

        :param value: String to set the stringCaseType parameter to.
        :returns: Instance of class mixed in.
        """
        possible_order_options = [
            "upper",
            "lower",
        ]
        if value not in possible_order_options:
            raise ValueError(
                f"stringCaseType must be one of {', '.join(possible_order_options)}"
            )
        return self._set(stringCaseType=value)

    def getStringCaseType(self) -> str:
        """
        Gets the stringCaseType parameter.

        :returns: String value of how to change the case of the string.
        """
        return self.getOrDefault(self.stringCaseType)


class StringCaseTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    StringCaseParams,
):
    """
    StringCaseLayer Spark Transformer for use in Spark pipelines.
    This transformer applies an upper, lower or capitalise operation
    on the input column.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        stringCaseType: Optional[str] = None,
    ) -> None:
        """
        Initializes an StringCaseTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param stringCaseType: How to change the case of the string. Must be one of:
        - 'upper'
        - 'lower'
        Default is 'lower'.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(stringCaseType="lower")
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which applies the given stringCaseType to the input column.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        string_case_type = self.getStringCaseType()

        def string_case(x: Column, case_type: str) -> Column:
            if case_type == "upper":
                return F.upper(x)
            elif case_type == "lower":
                return F.lower(x)
            else:
                raise ValueError(
                    f"""stringCaseType must be one of 'upper' or 'lower'.
                    Got {case_type}"""
                )

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: string_case(x, string_case_type),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the StringCaseLayer transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs the string casing operation.
        """
        return StringCaseLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            string_case_type=self.getStringCaseType(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringCaseLayer(BaseLayer):
    """
    Performs a string case transform on the input tensor.
    Supported string case types are 'upper' and 'lower'.
    """

    def __init__(
        self,
        string_case_type: str = "lower",
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the StringCaseLayer layer.

        :param string_case_type: The type of string case transform to perform.
        Supported types are 'upper' and 'lower'. Defaults to 'lower'.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.string_case_type = string_case_type

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the string case transform on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input is a
        single tensor. Raises an error if multiple tensors are passed in as an iterable.

        :param inputs: Input tensor to perform the string case transform on.
        :returns: The input tensor with the string case transform applied.
        """
        if self.string_case_type == "upper":
            return tf.strings.upper(inputs)
        elif self.string_case_type == "lower":
            return tf.strings.lower(inputs)
        else:
            raise ValueError(
                f"""stringCaseType must be one of 'upper' or 'lower'.
                Got {self.string_case_type}"""
            )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringCase layer.
        Used for saving and loading from a model.

        Specifically adds the `string_case_type` value to the configuration.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"string_case_type": self.string_case_type})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import MultiInputSingleOutputParams
from kamae.spark.utils import multi_input_single_output_array_transform
from kamae.tensorflow.layers import CosineSimilarityLayer

from .base import BaseTransformer


class CosineSimilarityTransformer(
    BaseTransformer,
    MultiInputSingleOutputParams,
):
    """
    Cosine Similarity Spark Transformer for use in Spark pipelines.
    This transformer computes the cosine similarity between two array columns.
    """

    @keyword_only
    def __init__(
        self,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes a CosineSimilarityTransformer transformer.

        :param inputCols: Input column names. Must be two columns.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def setInputCols(self, value: List[str]) -> "CosineSimilarityTransformer":
        """
        Sets the inputCols parameter. Ensures that there are only two input columns.

        :param value: List of input column names.
        :returns: Class instance.
        """
        if len(value) != 2:
            raise ValueError(
                f"""Expected 2 input columns, received {len(value)}
                input columns instead."""
            )
        return self._set(inputCols=value)

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which is the cosine similarity between the two input columns.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_col_names = self.getInputCols()
        input_cols = [F.col(c) for c in input_col_names]

        # Check both columns are arrays.
        datatypes = [
            self.get_column_datatype(dataset=dataset, column_name=input_col_name)
            for input_col_name in input_col_names
        ]
        if not all([isinstance(datatype, ArrayType) for datatype in datatypes]):
            raise TypeError(
                f"""Expected input columns to be of type ArrayType,
                received {datatypes} instead."""
            )

        # Compute dot product and the norms of the two arrays. The arrays are
        # represented by the "input_0" and "input_1" elements in the zipped array.
        def dot_product(x: Column) -> Column:
            return F.aggregate(
                x,
                F.lit(0.0),
                lambda acc, y: acc + y[input_col_names[0]] * y[input_col_names[1]],
            )

        def norm(x: Column, col_name: str) -> Column:
            return F.sqrt(
                F.aggregate(
                    x, F.lit(0.0), lambda acc, y: acc + y[col_name] * y[col_name]
                )
            )

        # If the norms are zero, then we match the tensorflow behavior and return 0.0.
        output_col = multi_input_single_output_array_transform(
            input_cols=input_cols,
            input_col_datatypes=datatypes,
            input_col_names=input_col_names,
            func=lambda x: F.coalesce(
                dot_product(x)
                / (norm(x, input_col_names[0]) * norm(x, input_col_names[1])),
                F.lit(0.0),
            ),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the cosine similarity transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         computes the cosine similarity between two arrays.
        """
        return CosineSimilarityLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            axis=-1,
            keepdims=True,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, IntegerType, StringType

from kamae.spark.params import HashIndexParams, SingleInputSingleOutputParams
from kamae.spark.utils import hash_udf, single_input_single_output_scalar_udf_transform
from kamae.tensorflow.layers import HashIndexLayer

from .base import BaseTransformer


class HashIndexTransformer(
    BaseTransformer,
    HashIndexParams,
    SingleInputSingleOutputParams,
):
    """
    Hash indexer Spark Transformer for use in Spark pipelines.
    This transformer hashes the input column and then bins it into
    the specified number of bins using modulo arithmetic.

    NOTE: If your data contains null characters:
    https://en.wikipedia.org/wiki/Null_character
    This transformer could fail since the hashing algorithm uses cannot accept null
    characters. If you have null characters in your data, you should remove them.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        numBins: Optional[int] = None,
        maskValue: Optional[str] = None,
    ) -> None:
        """
        Instantiates a HashIndexTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param numBins: Number of bins to use for hash indexing.
        :param maskValue: Mask value to use for hash indexing.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(maskValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column named outputCol with the
        hash indexed input column.

        :param dataset: Pyspark DataFrame to transform.
        :returns: Transformed pyspark dataFrame.
        """
        num_bins = self.getNumBins()
        mask_value = self.getMaskValue()

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_udf_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: hash_udf(label=x, num_bins=num_bins, mask_value=mask_value),
            udf_return_element_datatype=IntegerType(),
        )

        return dataset.withColumn(
            self.getOutputCol(),
            output_col,
        )

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the hash indexing.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the hash indexing operation.
        """
        return HashIndexLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            num_bins=self.getNumBins(),
            mask_value=self.getMaskValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input
from kamae.utils import get_condition_operator

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class BinLayer(BaseLayer):
    """
    Performs a binning operation on a given input tensor.

    The binning operation is performed by comparing the input tensor to a list of
    values using a list of operators. The bin label corresponding to the first
    condition that evaluates to True is returned.

    If no conditions evaluate to True, the default label is returned.
    """

    def __init__(
        self,
        condition_operators: List[str],
        bin_values: List[float],
        bin_labels: List[Union[float, int, str]],
        default_label: Union[float, int, str],
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initializes the BinLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param condition_operators: List of operators to use in the if statement.
        Can be one of:
            - "eq": Equal to
            - "neq": Not equal to
            - "lt": Less than
            - "leq": Less than or equal to
            - "gt": Greater than
            - "geq": Greater than or equal to
        :param bin_values: List of values to compare the input tensor to. Must be the
        same length as condition_operators.
        :param bin_labels: List of labels to use for each bin. Must be the same length
        as condition_operators.
        :param default_label: Label to use if none of the conditions are met.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if len(condition_operators) != len(bin_labels) != len(bin_values):
            raise ValueError(
                f"""condition_operators, bin_labels and bin_values must be the same
                length. Got lengths: {len(condition_operators)}, {len(bin_labels)},
                {len(bin_values)}"""
            )
        self.condition_operators = condition_operators
        self.bin_values = bin_values
        self.bin_labels = bin_labels
        self.default_label = default_label

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.int8,
            tf.uint8,
            tf.int16,
            tf.uint16,
            tf.int32,
            tf.uint32,
            tf.int64,
            tf.uint64,
        ]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs a binning operation on a given input tensor.

        Creates a string tensor of the same shape as the input tensor, where each
        element is the label of the bin that the corresponding element in the input
        tensor belongs to. The bin labels are determined by successively applying
        the condition operators to the input tensor, and returning the label of the
        first bin that the element belongs to.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Tensor to perform the binning operation on.
        :returns: The binned input tensor.
        """
        cond_op_fns = [get_condition_operator(op) for op in self.condition_operators]

        # Build default output tensor
        outputs = tf.constant(self.default_label)

        # Loop through the conditions.
        # Reverse the list of conditions so that we start from the last condition
        # and work backwards. This ensures that the first condition that is met
        # is the one that is used.
        conds = zip(cond_op_fns[::-1], self.bin_values[::-1], self.bin_labels[::-1])

        for cond_op, value, label in conds:
            # Ensure that the inputs and value are compatible dtypes
            cast_input, cast_value = self._force_cast_to_compatible_numeric_type(
                inputs, value
            )
            outputs = tf.where(
                cond_op(
                    cast_input,
                    cast_value,
                ),
                tf.constant(label),
                outputs,
            )

        return outputs

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the Bin layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "condition_operators": self.condition_operators,
                "bin_values": self.bin_values,
                "bin_labels": self.bin_labels,
                "default_label": self.default_label,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from pyspark.ml.param import Param, Params, TypeConverters


class HasLayerName(Params):
    """
    Mixin class for a layer name.
    """

    layerName = Param(
        Params._dummy(),
        "layerName",
        "Name of the layer",
        typeConverter=TypeConverters.toString,
    )

    def getLayerName(self) -> str:
        """
        Gets the value of the layerName parameter.

        :returns: Layer name.
        """
        return self.getOrDefault(self.layerName)
